#pragma once
#include "H5Cpp.h"
#include "aare/defs.hpp"
#include <filesystem>
#include <fmt/format.h>
#include <fstream>
#include <optional>



namespace aare {

using ns = std::chrono::nanoseconds;
/**
 * @brief Implementation used in Hdf5MasterFile to parse the file name
 */
class Hdf5FileNameComponents {
    bool m_old_scheme{false};
    std::filesystem::path m_base_path{};
    std::string m_base_name{};
    std::string m_ext{};
    int m_file_index{}; 

  public:
    Hdf5FileNameComponents(const std::filesystem::path &fname);

    /// @brief Get the filename including path of the master file.
    /// (i.e. what was passed in to the constructor))
    std::filesystem::path master_fname() const;

    /// @brief Get the filename including path of the data file.
    /// @param mod_id module id run_d[module_id]_f0_0
    /// @param file_id file id run_d0_f[file_id]_0
    std::filesystem::path data_fname(size_t mod_id, size_t file_id) const;

    const std::filesystem::path &base_path() const;
    const std::string &base_name() const;
    const std::string &ext() const;
    int file_index() const;
    void set_old_scheme(bool old_scheme);
};


/**
 * @brief Class for parsing a master file either in our .json format or the old
 * .Hdf5 format
 */
class Hdf5MasterFile {
    
    Hdf5FileNameComponents m_fnc;
    std::string m_version;
    DetectorType m_type;
    TimingMode m_timing_mode;
    xy m_geometry{};
    size_t m_image_size_in_bytes{};
    size_t m_pixels_y{};
    size_t m_pixels_x{};
    size_t m_max_frames_per_file{};
    FrameDiscardPolicy m_frame_discard_policy{};
    size_t m_frame_padding{};
    ScanParameters m_scan_parameters;
    size_t m_total_frames_expected{};
    std::optional<ns> m_exptime{};
    std::optional<ns> m_period{};
    std::optional<BurstMode> m_burst_mode;
    std::optional<size_t> m_number_of_udp_interfaces;
    size_t m_bitdepth{};
    std::optional<size_t> m_ten_giga;
    // thresholdenergy
    // thresholdall energy
    std::optional<ns> m_subexptime{};
    std::optional<ns> m_subperiod{};
    std::optional<uint8_t> m_quad;
    std::optional<size_t> m_number_of_rows;
    // ratecorr
    uint32_t m_adc_mask{}; // TODO! implement reading
    uint8_t m_analog_flag{};
    std::optional<size_t> m_analog_samples;
    uint8_t m_digital_flag{};
    std::optional<size_t> m_digital_samples;
    // dbitoffset
    // dbitlist
    // transceiver mask
    uint8_t m_transceiver_flag{};
    std::optional<size_t> m_transceiver_samples;
    // g1 roi
    std::optional<ROI> m_roi;
    // counter mask
    // exptimearray
    // gatedelay array
    // gates
    // additional json header
    size_t m_frames_in_file{};




    // TODO! should these be bool?




  public:
    Hdf5MasterFile(const std::filesystem::path &fpath);

    std::filesystem::path master_fname() const;
    std::filesystem::path data_fname(size_t mod_id, size_t file_id) const;

    const std::string &version() const; //!< For example "7.2"
    const DetectorType &detector_type() const;
    const TimingMode &timing_mode() const;
    xy geometry() const;
    size_t image_size_in_bytes() const;
    size_t pixels_y() const;
    size_t pixels_x() const;
    size_t max_frames_per_file() const;
    const FrameDiscardPolicy &frame_discard_policy() const;
    size_t frame_padding() const;
    ScanParameters scan_parameters() const;
    size_t total_frames_expected() const;
    std::optional<ns> exptime() const;
    std::optional<ns> period() const;
    std::optional<BurstMode> burst_mode() const;
    std::optional<size_t> number_of_udp_interfaces() const;
    size_t bitdepth() const;
    std::optional<size_t> ten_giga() const;
    // thresholdenergy
    // thresholdall energy
    std::optional<ns> subexptime() const;
    std::optional<ns> subperiod() const;
    std::optional<uint8_t> quad() const;
    std::optional<size_t> number_of_rows() const;
    // ratecorr
    std::optional<uint32_t> adc_mask() const; // TODO! implement reading
    std::optional<uint8_t> analog_flag() const;
    std::optional<size_t> analog_samples() const;
    std::optional<uint8_t> digital_flag() const;
    std::optional<size_t> digital_samples() const;
    // dbitoffset
    // dbitlist
    // transceiver mask    
    std::optional<uint8_t> transceiver_flag() const;
    std::optional<size_t> transceiver_samples() const;
    // g1 roi
    std::optional<ROI> roi() const;
    // counter mask
    // exptimearray
    // gatedelay array
    // gates
    // additional json header
    size_t frames_in_file() const;
    size_t n_modules() const;


  private:
    static const std::string metadata_group_name;
    void parse_acquisition_metadata(const std::filesystem::path &fpath);

    template <typename T>
    T h5_read_scalar_dataset(const H5::DataSet &dataset,
                             const H5::DataType &data_type);

    template <typename T>
    T h5_get_scalar_dataset(const H5::H5File &file,
                            const std::string &dataset_name);
};

template <>
std::string Hdf5MasterFile::h5_read_scalar_dataset<std::string>(
    const H5::DataSet &dataset, const H5::DataType &data_type);
} // namespace aare